# frozen_string_literal: true

require "eventmachine"
require "fiber"
require "promise"

class EMPromise < Promise
	class Trampoline
		attr_accessor :ready

		def initialize
			@ready = true
			@fiber = Fiber.new do |blk|
				Thread.current[:_em_promise_trampoline] = self
				loop do
					result = blk.call
					self.ready = true
					blk = Fiber.yield result
					self.ready = false
				end
			end
		end

		def submit(&blk)
			EM.next_tick do
				if ready
					@fiber.resume(blk)
				else
					submit(&blk)
				end
			end
		end
	end

	# Make sure that self.class.new inherits the trampoline
	def class
		tramp = @trampoline
		klass = super.dup
		klass.define_singleton_method(:new) do |*args|
			super(*args, trampoline: tramp)
		end
		klass
	end

	def initialize(deferrable=nil, trampoline: nil)
		super()
		@trampoline =
			trampoline ||
			Thread.current[:_em_promise_trampoline] ||
			Trampoline.new
		fulfill(deferrable) if deferrable
	end

	def fulfill(value, bind_defer=true)
		if bind_defer && value.is_a?(EM::Deferrable)
			value.callback { |x| fulfill(x, false) }
			value.errback(&method(:reject))
		else
			super(value)
		end
	end

	def defer
		@trampoline.submit { yield }
	end

	def wait
		fiber = Fiber.current
		resume = proc do
			EM.next_tick { fiber.resume }
		end

		self.then(resume, resume)

		# We might be in a trampoline, so keep that going
		tramp = Thread.current[:_em_promise_trampoline]
		raise "Trampoline already ready?" if tramp&.ready

		result = nil
		loop do
			tramp&.ready = true
			blk = Fiber.yield result
			tramp&.ready = false
			result = blk&.call
			break unless blk
		end
	end

	def self.reject(e)
		new.tap { |promise| promise.reject(e) }
	end

	def catch_only(*klasses, filter: ->(*) { false })
		catch do |e|
			if klasses.find { |klass| e.is_a?(klass) } || filter[e]
				yield e
			else
				EMPromise.reject(e)
			end
		end
	end

	def self.all(enumerable)
		super(enumerable.map { |input|
			if input.respond_to?(:promise)
				input.promise
			else
				input
			end
		})
	end
end

module EventMachine
	module Deferrable
		def promise
			EMPromise.new(self)
		end

		[:then, :rescue, :catch, :sync].each do |method|
			define_method(method) do |*args, &block|
				promise.public_send(method, *args, &block)
			end
		end
	end
end
